#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <conio.h>

typedef unsigned char u8;
typedef signed char s8;

#ifdef __BORLANDC__
typedef unsigned int u16;
typedef signed int s16;
#else
typedef unsigned short u16;
typedef signed short s16;
#endif

typedef unsigned long u32;
typedef signed long s32;

#ifdef __BORLANDC__
u8 far *SCR = (u8 far *)MK_FP(0xB800, 0);	// Borland C-style
#else
u8 *SCR = (u8 *)0xB8000;					// Watcom C-style
#endif

enum
{
	VIDEO_ERROR_NONE = 0,
	VIDEO_ERROR_OUT_OF_MEMORY,
	VIDEO_ERROR_VIDEOCARD_NOT_DETECTED,
	VIDEO_ERROR_BIOS_CALL_FAILED,
	VIDEO_ERROR_BIOS_NOT_DETECTED,
	VIDEO_ERROR_BIOS_VERSION_TOO_LOW,
	VIDEO_ERROR_NO_MODE_INFORMATION,
	VIDEO_ERROR_SET_VIDEO_MODE_FAILED,

	VIDEO_ERROR_MAX
};

static const char *SpectrumPaletteName[8] =
{
	" BLACK  ",
	"  BLUE  ",
	"  RED   ",
	"MAGENTA ",
	" GREEN  ",
	"  CYAN  ",
	" YELLOW ",
	" WHITE  "
};

static u8 const SpectrumPaletteRemap[2][16] =
{
	{
		0,  2,  4,  6,  1,  3,  8, 10,
		5,  7, 12, 14,  9, 11, 13, 15
	},
	{
		0,  4,  8, 12,  3,  6,  1, 10,
		5,  2,  9, 12,  3,  7, 11, 15
	}
};

static u8 bInitialScreenMode;

char *VIDEO_SetErrorString(u8 bReason)
{
	switch(bReason)
	{
		case VIDEO_ERROR_NONE:
			return "";
		case VIDEO_ERROR_OUT_OF_MEMORY:
			return "Video: Couldn't allocate enough memory for a back buffer";
		case VIDEO_ERROR_VIDEOCARD_NOT_DETECTED:
			return "Video: Videocard not detected";
		case VIDEO_ERROR_BIOS_CALL_FAILED:
			return "Video: BIOS call failed";
		case VIDEO_ERROR_BIOS_NOT_DETECTED:
			return "Video: BIOS not detected";
		case VIDEO_ERROR_BIOS_VERSION_TOO_LOW:
			return "Video: BIOS version too low";
		case VIDEO_ERROR_NO_MODE_INFORMATION:
			return "Video: BIOS could not provide mode information";
		case VIDEO_ERROR_SET_VIDEO_MODE_FAILED:
			return "Video: BIOS could not set video mode";
		default:
			// Generic error
			return "Video: Could not detect or initialise screen mode";
	}
}

void lower_it(char *stringy)
{
	size_t pos, x;

	pos = strlen(stringy);
	for(x = 0; x < pos; x++)
	{
		stringy[x] = tolower(stringy[x]);
	}
}

u8 HV_GetVideoMode(void)
{
	union REGS R;

	R.h.ah = 0x0F;
#ifdef __BORLANDC__
	int86(0x10, &R, &R);
#else
	int386(0x10, &R, &R);
#endif

	printf("Current Video Mode: 0x%02X\n", R.h.al);
	return R.h.al;
}

u8 HV_SetVideoMode(u8 mode, u8 *p_bReason)
{
	union REGS R;

#ifdef __BORLANDC__
	R.x.ax = mode;
	int86(0x10, &R, &R);
#else
	R.w.ax = mode;
	int386(0x10, &R, &R);
#endif

	//printf("Mode: %2X (Ret = %02X)\n", mode, R.h.al);

	// 1 = Success (mode requested == current mode)
	// 0 = Fail
	if(mode != HV_GetVideoMode())
	{
		if(p_bReason) *p_bReason = VIDEO_ERROR_SET_VIDEO_MODE_FAILED;
		return 0;
	}
	return 1;
}

void HV_StoreScreenMode(void)
{
	bInitialScreenMode = HV_GetVideoMode();
}

void DisplayBar(int x, int y, int height, int col)
{
	u8 far *base;

	// 640x200x2 = 80 bytes * 200 = 16,000 bytes
	// CGA interleaving applies... (so divide Y by 2)
	base = SCR + (x >> 2) + ((y >> 1) * 80);

	// Halve the height to simplify the odd/even lines
	height >>= 1;

	while(height)
	{
		// (even line)
		// Clearly, I cannot remember how to use far pointers in the Small Model under Borland C...
		//memset(base, (col | (col << 4)), 0x05);		// 5 bytes wide = 10 (chunky) pixels
		base[0] = base[1] = base[2] = base[3] = base[4] = (col | (col << 4));

		// + 8K (odd line)
		base += 0x2000;
		//memset(base, (col | (col << 4)), 0x05);		// 5 bytes wide = 10 (chunky) pixels
		base[0] = base[1] = base[2] = base[3] = base[4] = (col | (col << 4));

		// - 8K + 80
		base -= 0x2000;
		base += 80;

		height--;
	}
}

void Testcard(void)
{
	u8 far *base, bRemapCol;
	int col, y;

	// 640x200x2 = 80 bytes * 200 = 16,000 bytes
	// CGA interleaving applies...
	gotoxy(1, 11);
	for(col = 0x00; col < 0x10; col++)
	{
		// x * 5, y, height, colour
		DisplayBar(col * 20, 0, 76, col);

		//gotoxy((col * 5) + 2, 11);
		//printf("%03d", col);
		printf(" %03d ", col);
	}

	//gotoxy(((col >> 1) * 10) + 2, 24);
	gotoxy(1, 24);
	for(col = 0x00; col < 0x10; col++)
	{
		// x * 5, y, height, colour
		bRemapCol = ((col >> 1) & 0x07) | ((col & 0x01) << 3);
		if(col == 1) bRemapCol = 0;				// Don't display "bright black"

		DisplayBar(col * 20, 104, 76, SpectrumPaletteRemap[0][bRemapCol]);

		if(!(col & 0x01))
		{
			//gotoxy(((col >> 1) * 10) + 2, 24);
			//printf("%s", SpectrumPaletteName[col >> 1]);
			printf(" %s ", SpectrumPaletteName[col >> 1]);
		}
	}
}




u8 parse_params(int argc, char *param[])
{
	char temp_str[500];
	//char dest_str_2[500];
	//char temp_str[HOST_MAXPATH];
	u8 arg_num;

	if(!argc) return 0;

	arg_num = 0;

	while(arg_num < argc)
	{
		if((param[arg_num][0] == '-') || (param[arg_num][0] == '/'))
		{
			strcpy(temp_str, (char *)param[arg_num] + 1);
			lower_it((char *)temp_str);
			if(!(strcmp(temp_str, "?")))
			{
				printf("CCTest V0.0 Help (it has not been written yet)\n");
				printf("\n");
				return 0;
			}
			//if(!(strcmp(temp_str, "ini")))
			//{
			//	arg_num++;
			//	strcpy(dest_str_2, (char *)param[arg_num]);
			//	RemoveQuotes(dest_str_2);

			//	GetFullPath(dest_str_2);
			//	strcpy(g.the_rom_path[ROM_PATH_INI], dest_str_2);
			//}
		}
		arg_num++;
	}

	return 1;
}

int main(int argc, char *param[])
{
	u8 bReason = VIDEO_ERROR_NONE;
	u8 bContinue = 1;

	if(argc)
	{
		bContinue = parse_params(argc, param);
	}

	if(bContinue)
	{
		// Do something
		HV_StoreScreenMode();
		//printf("Current screen mode: 0x%02X\n", bInitialScreenMode);

		HV_SetVideoMode(0x06, &bReason);		// CGA 640x200

		if(bReason == VIDEO_ERROR_NONE)
		{
			// Composite, turn on colourburst
			outp(0x3D8, 0x10 | 0x08 | (0x04 & 0x00) | 0x02);	// Hi-res Graphics, Enable Video, Colour, Graphics Mode

			// Display testcard
			Testcard();

			// Wait for key
			while(!kbhit());
			getch();

			// Back to normal
			HV_SetVideoMode(bInitialScreenMode, NULL);
		}

		printf("%s\n", VIDEO_SetErrorString(bReason));
	}

	return 1;
}
